/*                                                                -*- mode: c; c-basic-offset: 8;  -*-
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 ============================================================================
 CLI used by the FTS cron job for generating the XML file (services.xml)
 ============================================================================
 */

#include "ServiceDiscoveryIfce.h"
#include "Definitions.h"
#include <stdlib.h> 

#define static /* empty */



/* FUNCTIONS ***********************************************************/
/**
 * Frees an SDVOList - there's no public interface to this in the API because
 * we never return a standalone VO list to the caller.
 *
 * @param  vos        VO list to free.
 *
 * @param  Nothing.
 */



/**********************************************************************
 * Global variables
 */

/* Usage message */
const char *tool_usage = "";

/* Help message */
const char *tool_help = "\t-h\t\tPrint this help text and exit.\n"
        "\t-q\t\tQuiet operation.\n"
        "\t-v\t\tBe more verbose.\n"
        "\t-n NAME\t\tName of the service to query\n"
        "\t-t TYPE\t\tType of service to query\n"
        "\t-s SITE\t\tSite of service to query\n"
        "\t--host|-H HOST\t\tHost of the service to query\n"
        "\t-e|--print-endpoint\t\tPrint endpoints only\n"
        "\t-N|--print-name\t\tPrint names only\n"
        "\t-x\t\tPrint details, too\n"
        "\t-a\t\tQuery associated services\n"
        "\t--xml|-X\t\tProduce xml output to stdout\n"
        "\t-d ATTR=VAL\tQuery services which match data\n";

/* ls-specific command line options */
const char *options = "hqvs:d:n:t:xeaNX";

/* Long options */
const struct option long_options[] ={
    { "xml", 0, NULL, 'X'},
    { "help", 0, NULL, 'h'},
    { "quiet", 0, NULL, 'q'},
    { "verbose", 0, NULL, 'v'},
    { "host", 1, NULL, 'H'},
    { "print-endpoint", 0, NULL, 'e'},
    { "print-name", 0, NULL, 'N'},
    { NULL, 0, NULL, 0}
};

/* Flag to turn on verbose output */
int verbose_flag;

/* Flag to request quiet operation */
int quiet_flag;

/* Name of the service to query */
char *name = NULL;

/* Site of the service to query */
char *site = NULL;

/* Type of the service to query */
char *type = NULL;

/* Host of the service to query */
char *host = NULL;

/* Set if print_details should be shown */
int print_details = 0;

/* Set if only endpoint is needed */
int print_endpoint = 0;

/* Set if only name is needed */
int print_name = 0;

/* Set if xml output is needed */
int print_xml = 0;

/* Set if associated services stuff is needed */
int associated;

/* Service Data list */
SDServiceDataList data_list;

/**********************************************************************
 * Logging functions
 */

static void info(const char *fmt, ...) {
    va_list ap;

    if (!verbose_flag)
        return;

    va_start(ap, fmt);
    vprintf(fmt, ap);
    printf("\n");
    fflush(stdout);
    va_end(ap);
}

static void error(const char *fmt, ...) {
    char *p, *msg;
    va_list ap;

    if (quiet_flag)
        return;
    if(!fmt)
	return;

    va_start(ap, fmt);
    msg = g_strdup_vprintf(fmt, ap);
    va_end(ap);

    /* If the command name is the same as the name of the failed method,
     * we may get duplication at the beginning of the message, like
     * mv: mv: Destination entry: "/foo" already exists!
     * Remove the duplication. */
    p = strchr(msg, ':');
    if (p) {
        int len = p - msg + 1;
        if (msg[len + 1] && !strncmp(msg, msg + len + 1, len))
            memmove(msg, msg + len + 1, strlen(msg) - len);
    }
    fprintf(stderr, "%s\n", msg);
    fflush(stderr);
    g_free(msg);
}

/**********************************************************************
 * Helpers
 */
static void uppercase(char *sPtr) {
    if (sPtr) {
        while (*sPtr != '\0') {
            *sPtr = toupper(*sPtr);
            sPtr++;
        }
    }
}

static void print_service_text(SDService *service) {

    if(!service)
    	return;

    if (print_endpoint) {
        printf("%s\n", service->endpoint);
    } else {
        printf("\n");
        printf("Name: %s\n", service->name);
        printf("Type: %s\n", service->type);
        printf("Endpoint: %s\n", service->endpoint);
        printf("Version: %s\n", service->version);
    }
}

static void print_service_xml(SDService *service) {

    if(!service)
    	return;

    printf("\n");
    printf("  <service name=\"%s\">\n", service->name);
    printf("    <parameters>\n");
    printf("      <endpoint>%s</endpoint>\n", service->endpoint);
    printf("      <type>%s</type>\n", service->type);
    printf("      <version>%s</version>\n", service->version);
    printf("    </parameters>\n");
    printf("  </service>\n");
}

static void print_service_details_xml(SDServiceDetails *serviceDetails) {

    if(!serviceDetails)
    	return;
	
    printf("\n");
    printf("  <service name=\"%s\">\n", serviceDetails->name);
    printf("    <parameters>\n");
    printf("      <endpoint>%s</endpoint>\n", serviceDetails->endpoint);
    printf("      <type>%s</type>\n", serviceDetails->type);
    printf("      <version>%s</version>\n", serviceDetails->version);
    uppercase(serviceDetails->site);
    printf("      <site>%s</site>\n", serviceDetails->site);
    if (serviceDetails->wsdl)
        printf("      <wsdl>%s</wsdl>\n", serviceDetails->wsdl);
    if (serviceDetails->administration)
        printf("      <administration>%s</administration>\n",
            serviceDetails->administration);
    if (serviceDetails->vos) {
        int i;
        printf("      <volist>\n");
        for (i = 0; i < serviceDetails->vos->numNames; i++)
            printf("        <vo>%s</vo>\n", serviceDetails->vos->names[i]);
        printf("      </volist>\n");
    }
    if (serviceDetails->associatedServices) {
        int i;
        printf("      <associatedservices>\n");
        for (i = 0; i < serviceDetails->associatedServices->numServices; i++)
            printf("        <name>%s</name>\n",
                serviceDetails->associatedServices->services[i]->name);
        printf("      </associatedservices>\n");
    }

    if (serviceDetails->data && serviceDetails->data->numItems) {
        int i;
        for (i = 0; i < serviceDetails->data->numItems; i++)
            printf("      <param name=\"%s\">%s</param>\n",
                serviceDetails->data->items[i].key,
                serviceDetails->data->items[i].value);
    }
    printf("    </parameters>\n");
    printf("  </service>\n");

}

static void print_service_details_text(SDServiceDetails *serviceDetails) {
    int i;
    SDServiceData *serviceData = NULL;
    
    if(!serviceDetails)
    	return;

    printf("\n");
    printf("Name: %s\n", serviceDetails->name);
    printf("Type: %s\n", serviceDetails->type);
    printf("Endpoint: %s\n", serviceDetails->endpoint);
    printf("Version: %s\n", serviceDetails->version);
    printf("Site: %s\n", serviceDetails->site);
    if (serviceDetails->wsdl)
        printf("WSDL: %s\n", serviceDetails->wsdl);
    if (serviceDetails->administration)
        printf("Administrator: %s\n", serviceDetails->administration);

    if (serviceDetails->vos) {
        printf("VOs: ");
        for (i = 0; i < serviceDetails->vos->numNames - 1; i++)
            printf("%s,", serviceDetails->vos->names[i]);
        printf("%s\n", serviceDetails->vos->names[serviceDetails->vos->numNames
                - 1]);
        printf("\n");
    }

    if (serviceDetails->associatedServices) {
        printf("Associated Services\n");
        for (i = 0; i < serviceDetails->associatedServices->numServices; i++) {
            printf("\t>\n");
            printf("\tName: %s\n",
                    serviceDetails->associatedServices->services[i]->name);
            printf("\tType: %s\n",
                    serviceDetails->associatedServices->services[i]->type);
            printf("\tEndpoint: %s\n",
                    serviceDetails->associatedServices->services[i]->endpoint);
            printf("\tVersion: %s\n",
                    serviceDetails->associatedServices->services[i]->version);
        }
    }

    if (serviceDetails->data && serviceDetails->data->numItems) {
        printf("Service Data:\n");
        for (i = 0; i < serviceDetails->data->numItems; i++) {
            serviceData = &(serviceDetails->data->items[i]);
            printf(" Key: %s - Value: %s\n", serviceData->key,
                    serviceData->value);
        }
    }
}

static void print_service(SDService *service) {
    SDException exc;
    SDServiceDetails *serviceDetails = NULL;
    
    if(!service)
    	return;

    if (print_details) {
        /* Need extra call for each service in the list to get details */
        serviceDetails = SD_getServiceDetails(service->name, &exc);

        if (exc.status != SDStatus_SUCCESS) {

            error(exc.reason);
            if (serviceDetails)
                SD_freeServiceDetails(serviceDetails);
            SD_freeException(&exc);
            return;
        } else {
           if(serviceDetails){
            if (print_xml) {
                print_service_details_xml(serviceDetails);
                SD_freeServiceDetails(serviceDetails);
            } else {
                print_service_details_text(serviceDetails);
                SD_freeServiceDetails(serviceDetails);
            }
	  }
        }
    } else {
        if (print_xml) {
            print_service_xml(service);
            
        } else {
            print_service_text(service);
            
        }
    }
   
}

static int print_service_list(SDServiceList *list) {
    int i;

    if (!list)
        return 1;

    for (i = 0; i < list->numServices; i++) {
        print_service(list->services[i]);
    }
    return 1;
}

/**********************************************************************
 * Tool implementation
 */

static int tool_doit(int argc G_GNUC_UNUSED, char *argv[] G_GNUC_UNUSED) {
    SDServiceList *list = NULL;
    SDService *service = NULL;
    SDException exc;
    G_GNUC_UNUSED SDVOList autovo;
    int i = 0;

    /* XXX Experimental VO autodetection */
    autovo.numNames = 0;
    autovo.names = NULL;

    /** Query by name */
    if (name) {
        service = SD_getService(name, &exc);
        if (exc.status != SDStatus_SUCCESS) {
            error(exc.reason);
            SD_freeService(service);
            SD_freeException(&exc);
            return EXIT_FAILURE;
        } else {
            print_service(service);
            SD_freeService(service);
        }

        /* print also the associated services */
        if (associated) {
            SDServiceList *list;
            list = SD_listAssociatedServices(name, type, NULL, NULL, &exc);
            if (exc.status != SDStatus_SUCCESS) {
                error(exc.reason);
                SD_freeServiceList(list);
                if (list)
                    SD_freeServiceList(list);
                SD_freeException(&exc);
                return EXIT_FAILURE;
            }
            if (!print_service_list(list)) {

                SD_freeServiceList(list);
                return EXIT_FAILURE;
            }
            SD_freeServiceList(list);
        }
    }        /** Query by type */
    else if (host) {
        list = SD_listServicesByHost(type, host, NULL, &exc);
        if (exc.status != SDStatus_SUCCESS) {
            error(exc.reason);
            SD_freeServiceList(list);
            SD_freeException(&exc);
            return EXIT_FAILURE;
        }
        if (!print_service_list(list))
            return EXIT_FAILURE;
        SD_freeServiceList(list);
    } else {

        /** List the services ... */
        if (!data_list.numItems) {
            list = SD_listServices(type, site, NULL, &exc);
            if (exc.status != SDStatus_SUCCESS) {
                error(exc.reason);
                SD_freeServiceList(list);
                SD_freeException(&exc);
                return EXIT_FAILURE;
            }
            if (!print_service_list(list))
                return EXIT_FAILURE;
              SD_freeServiceList(list);
        } else {
            /* With a data list */
            if (associated) {

                list = SD_listServicesByData(&data_list, type, site, NULL,
                        &exc);
                if (exc.status != SDStatus_SUCCESS) {
                    error("No services match the given data list");
                    if (list)
                        SD_freeServiceList(list);
                    SD_freeException(&exc);
                    return EXIT_FAILURE;
                }
                for (i = 0; i < list->numServices; i++) {
                    SDServiceList *list2;
                    list2 = SD_listAssociatedServices(
                            list->services[i]->name, type, NULL, NULL, &exc);
                    if (exc.status != SDStatus_SUCCESS) {

                        error(exc.reason);
                        SD_freeServiceList(list);
                        if (list2)
                            SD_freeServiceList(list2);
                        SD_freeException(&exc);
                        return EXIT_FAILURE;
                    }
                    if (!print_service_list(list2)) {

                        SD_freeServiceList(list);
                        return EXIT_FAILURE;
                    }
                    SD_freeServiceList(list2);
                }
            } else {

                list = SD_listServicesByData(&data_list, type, site, NULL,
                        &exc);
                if (exc.status != SDStatus_SUCCESS) {

                    error(exc.reason);
                    if (list)
                        SD_freeServiceList(list);
                    SD_freeException(&exc);
                    return EXIT_FAILURE;
                }
                if (!print_service_list(list))
                    return EXIT_FAILURE;
            }
            SD_freeServiceList(list);
        }
    }

    return EXIT_SUCCESS;
}

/**********************************************************************
 * Main routine
 */

int main(int argc, char *argv[]) {
    char *prog_name = NULL;
    char *attr = NULL;
    char *val = NULL;
    char *arg = NULL;
    int ret = 0;
    int c = 0;

    /*cli must use only and always bdii*/
    setenv("GLITE_SD_PLUGIN", "bdii", 1);

    /* Determine the program name we were called as */
    prog_name = strrchr(argv[0], '/');
    if (!prog_name)
        prog_name = argv[0];
    else
        prog_name++;

    /* Parse the command line options */
    while ((c = getopt_long(argc, argv, options, long_options, NULL)) != -1)
        switch (c) {
            case 'h':
                printf("Usage: %s [options]", prog_name);
                if (tool_usage)
                    printf(" %s\n", tool_usage);
                else
                    printf("\n");

                printf("Available options:\n%s", tool_help);
                exit(EXIT_SUCCESS);
            case 'q':
                quiet_flag++;
                verbose_flag = 0;
                break;
            case 'v':
                verbose_flag++;
                quiet_flag = 0;
                break;
            case 'd':
                arg = g_strdup(optarg);
                if (!arg)
                    return -1;
                attr = strtok(arg, "=");
                val = strtok(NULL, "=");
                if (!attr || !val) {
                    error("Attribute and/or value is missing for '-d' option!\n");
                    g_free(arg);
                    break;
                }
                data_list.items = realloc(data_list.items, (data_list.numItems + 1)
                        * sizeof (*data_list.items));
                if (!data_list.items) {
                    error("Out of memory error.");
                    g_free(arg);
                    exit(EXIT_FAILURE);
                }
                data_list.items[data_list.numItems].key = g_strdup(attr);
                data_list.items[data_list.numItems].value = g_strdup(val);
                data_list.numItems++;
                g_free(arg);
                break;
            case 'n':
                name = g_strdup(optarg);
                break;
            case 't':
                type = g_strdup(optarg);
                break;
            case 's':
                site = g_strdup(optarg);
                break;
            case 'x':
                print_details = 1;
                print_endpoint = 0;
                break;
            case 'e':
                print_details = 0;
                print_endpoint = 1;
                print_xml = 0;
                break;
            case 'H':
                host = g_strdup(optarg);
                break;
            case 'X':
                print_endpoint = 0;
                print_xml = 1;
                break;
            case 'a':
                associated = 1;
                break;
            case ':':
                error("Option argument is missing\n");
                exit(EXIT_FAILURE);
            case '?':
                error("Unknown command line option\n");
                exit(EXIT_FAILURE);
            default:
                error("Illegal command line arguments\n");
                exit(EXIT_FAILURE);
        }

    /* Drop the processed options from the argument list */
    argc -= optind;
    argv += optind;

    /* Perform the operation */
    ret = tool_doit(argc, argv);
   
    return ret;
}

